@@ -465,6 +465,14 @@ static int mbox_open_mailbox_append (CONTEXT *ctx, int flags)
 
 static int mbox_close_mailbox (CONTEXT *ctx)
 {
+  if (ctx->append)
+  {
+    mx_unlock_file (ctx->path, fileno (ctx->fp), 1);
+    mutt_unblock_signals ();
+  }
+
+  safe_fclose (&ctx->fp);
+
   return 0;
 }
 
@@ -794,7 +802,7 @@ void mbox_reset_atime (CONTEXT *ctx, struct stat *st)
  *	0	success
  *	-1	failure
  */
-int mbox_sync_mailbox (CONTEXT *ctx, int *index_hint)
+static int mbox_sync_mailbox (CONTEXT *ctx, int *index_hint)
 {
   char tempfile[_POSIX_PATH_MAX];
   char buf[32];
@@ -809,6 +817,7 @@ int mbox_sync_mailbox (CONTEXT *ctx, int *index_hint)
   FILE *fp = NULL;
   progress_t progress;
   char msgbuf[STRING];
+  BUFFY *tmp = NULL;
 
   /* sort message by their position in the mailbox on disk */
   if (Sort != SORT_ORDER)
@@ -1106,6 +1115,13 @@ int mbox_sync_mailbox (CONTEXT *ctx, int *index_hint)
   unlink (tempfile); /* remove partial copy of the mailbox */
   mutt_unblock_signals ();
 
+  if (option(OPTCHECKMBOXSIZE))
+  {
+    tmp = mutt_find_mailbox (ctx->path);
+    if (tmp && tmp->new == 0)
+      mutt_update_mailbox (tmp);
+  }
+
   return (0); /* signal success */
 
 bail:  /* Come here in case of disaster */
@@ -1182,6 +1198,7 @@ int mutt_reopen_mailbox (CONTEXT *ctx, int *index_hint)
     hash_destroy (&ctx->id_hash, NULL);
   if (ctx->subj_hash)
     hash_destroy (&ctx->subj_hash, NULL);
+  hash_destroy (&ctx->label_hash, NULL);
   mutt_clear_threads (ctx);
   FREE (&ctx->v2r);
   if (ctx->readonly)
@@ -1209,6 +1226,7 @@ int mutt_reopen_mailbox (CONTEXT *ctx, int *index_hint)
   ctx->changed = 0;
   ctx->id_hash = NULL;
   ctx->subj_hash = NULL;
+  mutt_make_label_hash (ctx);
 
   switch (ctx->magic)
   {
@@ -1346,6 +1364,7 @@ struct mx_ops mx_mbox_ops = {
   .commit_msg = mbox_commit_message,
   .open_new_msg = mbox_open_new_message,
   .check = mbox_check_mailbox,
+  .sync = mbox_sync_mailbox,
 };
 
 struct mx_ops mx_mmdf_ops = {
@@ -1357,4 +1376,5 @@ struct mx_ops mx_mmdf_ops = {
   .commit_msg = mmdf_commit_message,
   .open_new_msg = mbox_open_new_message,
   .check = mbox_check_mailbox,
+  .sync = mbox_sync_mailbox,
 };
