@@ -25,6 +25,7 @@
 #include <openssl/x509v3.h>
 #include <openssl/err.h>
 #include <openssl/rand.h>
+#include <openssl/evp.h>
 
 #undef _
 
@@ -54,6 +55,8 @@ static int entropy_byte_count = 0;
 #define HAVE_ENTROPY()	(!access(DEVRANDOM, R_OK) || entropy_byte_count >= 16)
 #endif
 
+/* index for storing hostname as application specific data in SSL structure */
+static int HostExDataIndex = -1;
 /* keep a handle on accepted certificates in case we want to
  * open up another connection to the same server in this session */
 static STACK_OF(X509) *SslSessionCerts = NULL;
@@ -78,12 +81,55 @@ static int tls_close (CONNECTION* conn);
 static void ssl_err (sslsockdata *data, int err);
 static void ssl_dprint_err_stack (void);
 static int ssl_cache_trusted_cert (X509 *cert);
-static int ssl_check_certificate (CONNECTION *conn, sslsockdata * data);
+static int ssl_verify_callback (int preverify_ok, X509_STORE_CTX *ctx);
 static int interactive_check_cert (X509 *cert, int idx, int len);
 static void ssl_get_client_cert(sslsockdata *ssldata, CONNECTION *conn);
 static int ssl_passwd_cb(char *buf, int size, int rwflag, void *userdata);
 static int ssl_negotiate (CONNECTION *conn, sslsockdata*);
 
+/* ssl certificate verification can behave strangely if there are expired
+ * certs loaded into the trusted store.  This function filters out expired
+ * certs.
+ * Previously the code used this form:
+ *     SSL_CTX_load_verify_locations (ssldata->ctx, SslCertFile, NULL);
+ */
+static int ssl_load_certificates (SSL_CTX *ctx)
+{
+  FILE *fp;
+  X509 *cert = NULL;
+  X509_STORE *store;
+  char buf[STRING];
+
+  dprint (2, (debugfile, "ssl_load_certificates: loading trusted certificates\n"));
+  store = SSL_CTX_get_cert_store (ctx);
+  if (!store)
+  {
+    store = X509_STORE_new ();
+    SSL_CTX_set_cert_store (ctx, store);
+  }
+
+  if ((fp = fopen (SslCertFile, "rt")) == NULL)
+    return 0;
+
+  while (NULL != PEM_read_X509 (fp, &cert, NULL, NULL))
+  {
+    if ((X509_cmp_current_time (X509_get_notBefore (cert)) >= 0) ||
+        (X509_cmp_current_time (X509_get_notAfter (cert)) <= 0))
+    {
+      dprint (2, (debugfile, "ssl_load_certificates: filtering expired cert: %s\n",
+              X509_NAME_oneline (X509_get_subject_name (cert), buf, sizeof (buf))));
+    }
+    else
+    {
+      X509_STORE_add_cert (store, cert);
+    }
+  }
+  X509_free (cert);
+  safe_fclose (&fp);
+
+  return 1;
+}
+
 /* mutt_ssl_starttls: Negotiate TLS over an already opened connection.
  *   TODO: Merge this code better with ssl_socket_open. */
 int mutt_ssl_starttls (CONNECTION* conn)
@@ -133,6 +179,18 @@ int mutt_ssl_starttls (CONNECTION* conn)
     goto bail_ctx;
   }
 
+  if (option (OPTSSLSYSTEMCERTS))
+  {
+    if (! SSL_CTX_set_default_verify_paths (ssldata->ctx))
+    {
+      dprint (1, (debugfile, "mutt_ssl_starttls: Error setting default verify paths\n"));
+      goto bail_ctx;
+    }
+  }
+
+  if (SslCertFile && !ssl_load_certificates (ssldata->ctx))
+    dprint (1, (debugfile, "mutt_ssl_starttls: Error loading trusted certificates\n"));
+
   ssl_get_client_cert(ssldata, conn);
 
   if (SslCiphers) {
@@ -370,6 +428,19 @@ static int ssl_socket_open (CONNECTION * conn)
     SSL_CTX_set_options(data->ctx, SSL_OP_NO_SSLv3);
   }
 
+  if (option (OPTSSLSYSTEMCERTS))
+  {
+    if (! SSL_CTX_set_default_verify_paths (data->ctx))
+    {
+      dprint (1, (debugfile, "ssl_socket_open: Error setting default verify paths\n"));
+      mutt_socket_close (conn);
+      return -1;
+    }
+  }
+
+  if (SslCertFile && !ssl_load_certificates (data->ctx))
+    dprint (1, (debugfile, "ssl_socket_open: Error loading trusted certificates\n"));
+
   ssl_get_client_cert(data, conn);
 
   if (SslCiphers) {
@@ -400,6 +471,19 @@ static int ssl_negotiate (CONNECTION *conn, sslsockdata* ssldata)
   int err;
   const char* errmsg;
 
+  if ((HostExDataIndex = SSL_get_ex_new_index (0, "host", NULL, NULL, NULL)) == -1)
+  {
+    dprint (1, (debugfile, "failed to get index for application specific data\n"));
+    return -1;
+  }
+
+  if (! SSL_set_ex_data (ssldata->ssl, HostExDataIndex, conn->account.host))
+  {
+    dprint (1, (debugfile, "failed to save hostname in SSL structure\n"));
+    return -1;
+  }
+
+  SSL_set_verify (ssldata->ssl, SSL_VERIFY_PEER, ssl_verify_callback);
   SSL_set_mode (ssldata->ssl, SSL_MODE_AUTO_RETRY);
 
   if ((err = SSL_connect (ssldata->ssl)) != 1)
@@ -422,17 +506,6 @@ static int ssl_negotiate (CONNECTION *conn, sslsockdata* ssldata)
     return -1;
   }
 
-  ssldata->cert = SSL_get_peer_certificate (ssldata->ssl);
-  if (!ssldata->cert)
-  {
-    mutt_error (_("Unable to get certificate from peer"));
-    mutt_sleep (1);
-    return -1;
-  }
-
-  if (!ssl_check_certificate (conn, ssldata))
-    return -1;
-
   /* L10N:
      %1$s is version (e.g. "TLSv1.2")
      %2$s is cipher_version (e.g. "TLSv1/SSLv3")
@@ -558,35 +631,24 @@ static void ssl_dprint_err_stack (void)
 }
 
 
-static char *x509_get_part (char *line, const char *ndx)
+static char *x509_get_part (X509_NAME *name, int nid)
 {
   static char ret[SHORT_STRING];
-  char *c, *c2;
 
-  strfcpy (ret, _("Unknown"), sizeof (ret));
-
-  c = strstr (line, ndx);
-  if (c)
-  {
-    c += strlen (ndx);
-    c2 = strchr (c, '/');
-    if (c2)
-      *c2 = '\0';
-    strfcpy (ret, c, sizeof (ret));
-    if (c2)
-      *c2 = '/';
-  }
+  if (!name ||
+      X509_NAME_get_text_by_NID (name, nid, ret, sizeof (ret)) < 0)
+    strfcpy (ret, _("Unknown"), sizeof (ret));
 
   return ret;
 }
 
-static void x509_fingerprint (char *s, int l, X509 * cert)
+static void x509_fingerprint (char *s, int l, X509 * cert, const EVP_MD *(*hashfunc)(void))
 {
   unsigned char md[EVP_MAX_MD_SIZE];
   unsigned int n;
   int j;
 
-  if (!X509_digest (cert, EVP_md5 (), md, &n))
+  if (!X509_digest (cert, hashfunc(), md, &n))
   {
     snprintf (s, l, _("[unable to calculate]"));
   }
@@ -619,61 +681,6 @@ static char *asn1time_to_string (ASN1_UTCTIME *tm)
   return buf;
 }
 
-static int check_certificate_by_signer (X509 *peercert)
-{
-  X509_STORE_CTX *xsc;
-  X509_STORE *ctx;
-  int pass = 0, i;
-
-  ctx = X509_STORE_new ();
-  if (ctx == NULL) return 0;
-
-  if (option (OPTSSLSYSTEMCERTS))
-  {
-    if (X509_STORE_set_default_paths (ctx))
-      pass++;
-    else
-      dprint (2, (debugfile, "X509_STORE_set_default_paths failed\n"));
-  }
-
-  if (X509_STORE_load_locations (ctx, SslCertFile, NULL))
-    pass++;
-  else
-    dprint (2, (debugfile, "X509_STORE_load_locations failed\n"));
-
-  for (i = 0; i < sk_X509_num (SslSessionCerts); i++)
-    pass += (X509_STORE_add_cert (ctx, sk_X509_value (SslSessionCerts, i)) != 0);
-
-  if (pass == 0)
-  {
-    /* nothing to do */
-    X509_STORE_free (ctx);
-    return 0;
-  }
-
-  xsc = X509_STORE_CTX_new();
-  if (xsc == NULL) return 0;
-  X509_STORE_CTX_init (xsc, ctx, peercert, SslSessionCerts);
-
-  pass = (X509_verify_cert (xsc) > 0);
-#ifdef DEBUG
-  if (! pass)
-  {
-    char buf[SHORT_STRING];
-    int err;
-
-    err = X509_STORE_CTX_get_error (xsc);
-    snprintf (buf, sizeof (buf), "%s (%d)",
-	X509_verify_cert_error_string(err), err);
-    dprint (2, (debugfile, "X509_verify_cert: %s\n", buf));
-  }
-#endif
-  X509_STORE_CTX_free (xsc);
-  X509_STORE_free (ctx);
-
-  return pass;
-}
-
 static int compare_certificates (X509 *cert, X509 *peercert,
   unsigned char *peermd, unsigned int peermdlen)
 {
@@ -741,7 +748,7 @@ static int check_certificate_by_digest (X509 *peercert)
     }
     if (X509_cmp_current_time (X509_get_notAfter (peercert)) <= 0)
     {
-      dprint (2, (debugfile, "Server certificate has expired"));
+      dprint (2, (debugfile, "Server certificate has expired\n"));
       mutt_error (_("Server certificate has expired"));
       mutt_sleep (2);
       return 0;
@@ -757,7 +764,7 @@ static int check_certificate_by_digest (X509 *peercert)
     return 0;
   }
 
-  while ((cert = PEM_read_X509 (fp, &cert, NULL, NULL)) != NULL)
+  while (PEM_read_X509 (fp, &cert, NULL, NULL) != NULL)
   {
     pass = compare_certificates (cert, peercert, peermd, peermdlen) ? 0 : 1;
 
@@ -919,107 +926,103 @@ static int ssl_cache_trusted_cert (X509 *c)
   return (sk_X509_push (SslSessionCerts, X509_dup(c)));
 }
 
-/* check whether cert is preauthorized. If host is not null, verify that
- * it matches the certificate.
- * Return > 0: authorized, < 0: problems, 0: unknown validity */
-static int ssl_check_preauth (X509 *cert, const char* host)
+/* certificate verification callback, called for each certificate in the chain
+ * sent by the peer, starting from the root; returning 1 means that the given
+ * certificate is trusted, returning 0 immediately aborts the SSL connection */
+static int ssl_verify_callback (int preverify_ok, X509_STORE_CTX *ctx)
 {
-  char buf[SHORT_STRING];
+  char buf[STRING];
+  const char *host;
+  int len, pos;
+  X509 *cert;
+  SSL *ssl;
+
+  if (! (ssl = X509_STORE_CTX_get_ex_data (ctx, SSL_get_ex_data_X509_STORE_CTX_idx ())))
+  {
+    dprint (1, (debugfile, "ssl_verify_callback: failed to retrieve SSL structure from X509_STORE_CTX\n"));
+    return 0;
+  }
+  if (! (host = SSL_get_ex_data (ssl, HostExDataIndex)))
+  {
+    dprint (1, (debugfile, "ssl_verify_callback: failed to retrieve hostname from SSL structure\n"));
+    return 0;
+  }
+
+  cert = X509_STORE_CTX_get_current_cert (ctx);
+  pos = X509_STORE_CTX_get_error_depth (ctx);
+  len = sk_X509_num (X509_STORE_CTX_get_chain (ctx));
+
+  dprint (1, (debugfile, "ssl_verify_callback: checking cert chain entry %s (preverify: %d)\n",
+              X509_NAME_oneline (X509_get_subject_name (cert),
+                                 buf, sizeof (buf)), preverify_ok));
 
   /* check session cache first */
   if (check_certificate_cache (cert))
   {
-    dprint (2, (debugfile, "ssl_check_preauth: using cached certificate\n"));
+    dprint (2, (debugfile, "ssl_verify_callback: using cached certificate\n"));
     return 1;
   }
 
+  /* check hostname only for the leaf certificate */
   buf[0] = 0;
-  if (host && option (OPTSSLVERIFYHOST) != MUTT_NO)
+  if (pos == 0 && option (OPTSSLVERIFYHOST) != MUTT_NO)
   {
     if (!check_host (cert, host, buf, sizeof (buf)))
     {
       mutt_error (_("Certificate host check failed: %s"), buf);
       mutt_sleep (2);
-      return -1;
+      return interactive_check_cert (cert, pos, len);
     }
-    dprint (2, (debugfile, "ssl_check_preauth: hostname check passed\n"));
+    dprint (2, (debugfile, "ssl_verify_callback: hostname check passed\n"));
   }
 
-  if (check_certificate_by_signer (cert))
+  if (!preverify_ok)
   {
-    dprint (2, (debugfile, "ssl_check_preauth: signer check passed\n"));
-    return 1;
-  }
+    /* automatic check from user's database */
+    if (SslCertFile && check_certificate_by_digest (cert))
+    {
+      dprint (2, (debugfile, "ssl_verify_callback: digest check passed\n"));
+      return 1;
+    }
 
-  /* automatic check from user's database */
-  if (SslCertFile && check_certificate_by_digest (cert))
-  {
-    dprint (2, (debugfile, "ssl_check_preauth: digest check passed\n"));
-    return 1;
-  }
-
-  return 0;
-}
-
-static int ssl_check_certificate (CONNECTION *conn, sslsockdata *data)
-{
-  int i, preauthrc, chain_len;
-  STACK_OF(X509) *chain;
-  X509 *cert;
 #ifdef DEBUG
-  char buf[STRING];
-
-  dprint (1, (debugfile, "ssl_check_certificate: checking cert %s\n",
-              X509_NAME_oneline (X509_get_subject_name (data->cert),
-                                 buf, sizeof (buf))));
+    /* log verification error */
+    {
+      int err = X509_STORE_CTX_get_error (ctx);
+      snprintf (buf, sizeof (buf), "%s (%d)",
+         X509_verify_cert_error_string (err), err);
+      dprint (2, (debugfile, "X509_verify_cert: %s\n", buf));
+    }
 #endif
 
-  if ((preauthrc = ssl_check_preauth (data->cert, conn->account.host)) > 0)
-    return preauthrc;
-
-  chain = SSL_get_peer_cert_chain (data->ssl);
-  chain_len = sk_X509_num (chain);
-  /* negative preauthrc means the certificate won't be accepted without
-   * manual override. */
-  if (preauthrc < 0 || !chain || (chain_len <= 1))
-    return interactive_check_cert (data->cert, 0, 0);
-
-  /* check the chain from root to peer. */
-  for (i = chain_len-1; i >= 0; i--)
-  {
-    cert = sk_X509_value (chain, i);
-
-    dprint (1, (debugfile, "ssl_check_certificate: checking cert chain entry %s\n",
-                X509_NAME_oneline (X509_get_subject_name (cert),
-                                   buf, sizeof (buf))));
-
-    /* if the certificate validates or is manually accepted, then add it to
-     * the trusted set and recheck the peer certificate */
-    if (ssl_check_preauth (cert, NULL)
-	|| interactive_check_cert (cert, i, chain_len))
-    {
-      ssl_cache_trusted_cert (cert);
-      if (ssl_check_preauth (data->cert, conn->account.host))
-	return 1;
-    }
+    /* prompt user */
+    return interactive_check_cert (cert, pos, len);
   }
 
-  return 0;
+  return 1;
 }
 
 static int interactive_check_cert (X509 *cert, int idx, int len)
 {
-  static const char * const part[] =
-    {"/CN=", "/Email=", "/O=", "/OU=", "/L=", "/ST=", "/C="};
+  static const int part[] =
+    { NID_commonName,             /* CN */
+      NID_pkcs9_emailAddress,     /* Email */
+      NID_organizationName,       /* O */
+      NID_organizationalUnitName, /* OU */
+      NID_localityName,           /* L */
+      NID_stateOrProvinceName,    /* ST */
+      NID_countryName             /* C */ };
+  X509_NAME *x509_subject;
+  X509_NAME *x509_issuer;
   char helpstr[LONG_STRING];
   char buf[STRING];
   char title[STRING];
   MUTTMENU *menu = mutt_new_menu (MENU_GENERIC);
   int done, row, i;
+  unsigned u;
   FILE *fp;
-  char *name = NULL, *c;
 
-  menu->max = 19;
+  menu->max = mutt_array_size (part) * 2 + 10;
   menu->dialog = (char **) safe_calloc (1, menu->max * sizeof (char *));
   for (i = 0; i < menu->max; i++)
     menu->dialog[i] = (char *) safe_calloc (1, SHORT_STRING * sizeof (char));
@@ -1027,25 +1030,18 @@ static int interactive_check_cert (X509 *cert, int idx, int len)
   row = 0;
   strfcpy (menu->dialog[row], _("This certificate belongs to:"), SHORT_STRING);
   row++;
-  name = X509_NAME_oneline (X509_get_subject_name (cert),
-			    buf, sizeof (buf));
-
-  for (i = 0; i < 5; i++)
-  {
-    c = x509_get_part (name, part[i]);
-    snprintf (menu->dialog[row++], SHORT_STRING, "   %s", c);
-  }
+  x509_subject = X509_get_subject_name (cert);
+  for (u = 0; u < mutt_array_size (part); u++)
+    snprintf (menu->dialog[row++], SHORT_STRING, "   %s",
+              x509_get_part (x509_subject, part[u]));
 
   row++;
   strfcpy (menu->dialog[row], _("This certificate was issued by:"), SHORT_STRING);
   row++;
-  name = X509_NAME_oneline (X509_get_issuer_name (cert),
-			    buf, sizeof (buf));
-  for (i = 0; i < 5; i++)
-  {
-    c = x509_get_part (name, part[i]);
-    snprintf (menu->dialog[row++], SHORT_STRING, "   %s", c);
-  }
+  x509_issuer = X509_get_issuer_name (cert);
+  for (u = 0; u < mutt_array_size (part); u++)
+    snprintf (menu->dialog[row++], SHORT_STRING, "   %s",
+              x509_get_part (x509_issuer, part[u]));
 
   row++;
   snprintf (menu->dialog[row++], SHORT_STRING, _("This certificate is valid"));
@@ -1056,8 +1052,11 @@ static int interactive_check_cert (X509 *cert, int idx, int len)
 
   row++;
   buf[0] = '\0';
-  x509_fingerprint (buf, sizeof (buf), cert);
-  snprintf (menu->dialog[row++], SHORT_STRING, _("Fingerprint: %s"), buf);
+  x509_fingerprint (buf, sizeof (buf), cert, EVP_sha1);
+  snprintf (menu->dialog[row++], SHORT_STRING, _("SHA1 Fingerprint: %s"), buf);
+  buf[0] = '\0';
+  x509_fingerprint (buf, sizeof (buf), cert, EVP_md5);
+  snprintf (menu->dialog[row++], SHORT_STRING, _("MD5 Fingerprint: %s"), buf);
 
   snprintf (title, sizeof (title),
 	    _("SSL Certificate check (certificate %d of %d in chain)"),
@@ -1122,6 +1121,7 @@ static int interactive_check_cert (X509 *cert, int idx, int len)
   }
   unset_option(OPTIGNOREMACROEVENTS);
   mutt_menuDestroy (&menu);
+  set_option (OPTNEEDREDRAW);
   dprint (2, (debugfile, "ssl interactive_check_cert: done=%d\n", done));
   return (done == 2);
 }
